use super::{File, FileSystemMap};
use std::path::PathBuf;
use byteorder::{ByteOrder, LittleEndian};
use no_comment::{IntoWithoutComments as _, languages};
use tool_helper::{Error, format_if_error, read_file};

pub type LBANameVec = Vec<String>;

#[repr(packed)]
struct OverlayHeader {
    pub start_adr: u32,
    pub _lba_count: u16,
}

impl OverlayHeader {
    pub fn lba_count_offset() -> usize {
        let dummy = OverlayHeader{start_adr: 0, _lba_count: 0}.start_adr;
        std::mem::size_of_val(&dummy)
    }
}

pub fn load_from(file_name: &str, file_path: PathBuf, lba_source: PathBuf) -> Result<File, Error> {
    let content      = load_content(&file_path)?;
    let lba_names    = load_lba_names(lba_source)?;
    let content_size = format_if_error!(tool_helper::compress::lz4(&content, 16), "Compressing {} failed with \"{error_text}\"", file_path.to_string_lossy())?.len();
    
    Ok(File::new_overlay(file_name, content, lba_names, content_size)?)
}

pub fn update_content(_content: Vec<u8>, lba_names: LBANameVec, file_map: &FileSystemMap) -> Result<Vec<u8>, Error> {
    for mut lba_name in lba_names {
        if lba_name.find(';').is_none() {
            lba_name.push_str(";1");
        }

        print!("Searching {}... ", lba_name);
        if let Some(info) = file_map.get(&lba_name) {
            println!("found @ {}", info.borrow().get_track_rel_lba())
        }

        else {
            println!("Not found...");
        }
    }
    
    Err(Error::not_implemented("update_overlay"))
}

fn load_content(file_path: &PathBuf) -> Result<Vec<u8>, Error> {
    let     overlay_header_size = std::mem::size_of::<OverlayHeader>();
    let mut content             = read_file(&file_path)?;

    if content.len() < overlay_header_size {
        return Err(Error::from_text(format!("Overlay {} has no header!", file_path.to_string_lossy())));
    }

    let lba_count       = LittleEndian::read_u16(&content[OverlayHeader::lba_count_offset()..]) as usize;
    let lba_header_size = lba_count*std::mem::size_of::<u32>();
    let lba_header      = &mut content[overlay_header_size..(overlay_header_size+lba_header_size)];

    if lba_header.is_empty() {
        return Err(Error::from_text(format!("LBA header of overlay {} is smaller then {} elements", file_path.to_string_lossy(), lba_count)));
    }

    let mut count = 0;
    for byte in lba_header {
        *byte = count as u8;
        count += 1;
    }

    Ok(content)
}

fn load_lba_names(lba_source: PathBuf) -> Result<LBANameVec, Error> {
    const LBA_DECLARATION:&'static str = "__jabyengine_request_lba_for";
    fn get_part_of_interest(file: String, lba_source: &PathBuf) -> Result<String, Error> {
        const LBA_START_POINT:&'static str = "__jabyengine_start_lba_request";
        const LBA_END_POINT:&'static str   = "__jabyengine_end_lba_request";

        let start = file.find(LBA_START_POINT).ok_or_else(|| {
            Error::from_text(format!("No LBA start found in file {}", lba_source.to_string_lossy()))
        })? + LBA_START_POINT.len();
    
        let end = file.find(LBA_END_POINT).ok_or_else(|| {
            Error::from_text(format!("No LBA end found in file {}", lba_source.to_string_lossy()))
        })?;

        Ok(file[start..end].to_owned())
    }

    let     file      = std::fs::read_to_string(&lba_source)?.chars().without_comments(languages::c()).collect::<String>();
    let     file      = get_part_of_interest(file, &lba_source)?;
    let mut lba_names = Vec::new();

    for pair in file.split(LBA_DECLARATION) {
        if let Some(mut start) = pair.find('"') {
            start += 1;
            if let Some(end) = pair[start..].find('"') {
                lba_names.push(pair[start..(start + end)].to_owned());
            }
        }
    }

    Ok(lba_names)
}